# -*- coding: utf-8 -*-
import platform
import sys


def user_agent(name, version, extras=None):
    """
    Returns an internet-friendly user_agent string.

    The majority of this code has been wilfully stolen from the equivalent
    function in Requests.

    :param name: The intended name of the user-agent, e.g. "python-requests".
    :param version: The version of the user-agent, e.g. "0.0.1".
    :param extras: List of two-item tuples that are added to the user-agent
        string.
    :returns: Formatted user-agent string
    :rtype: str
    """
    try:
        p_system = platform.system()
        p_release = platform.release()
    except IOError:
        p_system = 'Unknown'
        p_release = 'Unknown'

    if extras is None:
        extras = []

    if any(len(extra) != 2 for extra in extras):
        raise ValueError('Extras should be a sequence of two item tuples.')

    format_string = '%s/%s'

    extra_pieces = [
        format_string % (extra_name, extra_version)
        for extra_name, extra_version in extras
    ]

    user_agent_pieces = ([format_string % (name, version)] + extra_pieces +
                         [_implementation_string(),
                          '%s/%s' % (p_system, p_release)])

    return " ".join(user_agent_pieces)


def _implementation_string():
    """
    Returns a string that provides both the name and the version of the Python
    implementation currently running. For example, on CPython 2.7.5 it will
    return "CPython/2.7.5".

    This function works best on CPython and PyPy: in particular, it probably
    doesn't work for Jython or IronPython. Future investigation should be done
    to work out the correct shape of the code for those platforms.
    """
    implementation = platform.python_implementation()

    if implementation == 'CPython':
        implementation_version = platform.python_version()
    elif implementation == 'PyPy':
        implementation_version = '%s.%s.%s' % (sys.pypy_version_info.major,
                                               sys.pypy_version_info.minor,
                                               sys.pypy_version_info.micro)
        if sys.pypy_version_info.releaselevel != 'final':
            implementation_version = ''.join([
                implementation_version, sys.pypy_version_info.releaselevel
                ])
    elif implementation == 'Jython':
        implementation_version = platform.python_version()  # Complete Guess
    elif implementation == 'IronPython':
        implementation_version = platform.python_version()  # Complete Guess
    else:
        implementation_version = 'Unknown'

    return "%s/%s" % (implementation, implementation_version)



"""Module containing tests for requests_toolbelt.threaded API."""

import mock
import pytest

from requests_toolbelt._compat import queue
from requests_toolbelt import threaded


def test_creates_a_pool_for_the_user():
    """Assert a Pool object is used correctly and as we expect.

    This just ensures that we're not jumping through any extra hoops with our
    internal usage of a Pool object.
    """
    mocked_pool = mock.Mock(spec=['join_all', 'responses', 'exceptions'])
    with mock.patch('requests_toolbelt.threaded.pool.Pool') as Pool:
        Pool.return_value = mocked_pool
        threaded.map([{}, {}])

    assert Pool.called is True
    _, kwargs = Pool.call_args
    assert 'job_queue' in kwargs
    assert isinstance(kwargs['job_queue'], queue.Queue)
    mocked_pool.join_all.assert_called_once_with()
    mocked_pool.responses.assert_called_once_with()
    mocked_pool.exceptions.assert_called_once_with()


def test_raises_a_value_error_for_non_dictionaries():
    """Exercise our lazy valdation."""
    with pytest.raises(ValueError):
        threaded.map([[], []])


def test_raises_a_value_error_for_falsey_requests():
    """Assert that the requests param is truthy."""
    with pytest.raises(ValueError):
        threaded.map([])

    with pytest.raises(ValueError):
        threaded.map(None)


def test_passes_on_kwargs():
    """Verify that we pass on kwargs to the Pool constructor."""
    mocked_pool = mock.Mock(spec=['join_all', 'responses', 'exceptions'])
    with mock.patch('requests_toolbelt.threaded.pool.Pool') as Pool:
        Pool.return_value = mocked_pool
        threaded.map([{}, {}], num_processes=1000,
                     initializer=test_passes_on_kwargs)

    _, kwargs = Pool.call_args
    assert 'job_queue' in kwargs
    assert 'num_processes' in kwargs
    assert 'initializer' in kwargs

    assert kwargs['num_processes'] == 1000
    assert kwargs['initializer'] == test_passes_on_kwargs



# -*- coding: utf-8 -*-
import requests
import unittest
import mock

from requests_toolbelt.auth.guess import GuessAuth, GuessProxyAuth
from . import get_betamax


class TestGuessAuth(unittest.TestCase):
    def setUp(self):
        self.session = requests.Session()
        self.recorder = get_betamax(self.session)

    def cassette(self, name):
        return self.recorder.use_cassette(
            'httpbin_guess_auth_' + name,
            match_requests_on=['method', 'uri', 'digest-auth']
        )

    def test_basic(self):
        with self.cassette('basic'):
            r = self.session.request(
                'GET', 'http://httpbin.org/basic-auth/user/passwd',
                auth=GuessAuth('user', 'passwd'))

        assert r.json() == {'authenticated': True, 'user': 'user'}

    def test_digest(self):
        with self.cassette('digest'):
            r = self.session.request(
                'GET', 'http://httpbin.org/digest-auth/auth/user/passwd',
                auth=GuessAuth('user', 'passwd'))

        assert r.json() == {'authenticated': True, 'user': 'user'}

    def test_no_auth(self):
        with self.cassette('none'):
            url = 'http://httpbin.org/get?a=1'
            r = self.session.request('GET', url,
                                     auth=GuessAuth('user', 'passwd'))

            j = r.json()
            assert j['args'] == {'a': '1'}
            assert j['url'] == url
            assert 'user' not in r.text
            assert 'passwd' not in r.text


class TestGuessProxyAuth(unittest.TestCase):

    @mock.patch('requests_toolbelt.auth.http_proxy_digest.HTTPProxyDigestAuth.handle_407')
    def test_handle_407_header_digest(self, mock_handle_407):
        r = requests.Response()
        r.headers['Proxy-Authenticate'] = 'Digest nonce="d2b19757d3d656a283c99762cbd1097b", opaque="1c311ad1cc6e6183b83bc75f95a57893", realm="me@kennethreitz.com", qop=auth'

        guess_auth = GuessProxyAuth(None, None, "user", "passwd")
        guess_auth.handle_407(r)

        mock_handle_407.assert_called_with(r)

    @mock.patch('requests.auth.HTTPProxyAuth.__call__')
    @mock.patch('requests.cookies.extract_cookies_to_jar')
    def test_handle_407_header_basic(self, extract_cookies_to_jar, proxy_auth_call):
        req = mock.Mock()
        r = mock.Mock()
        r.headers = dict()
        r.request.copy.return_value = req

        proxy_auth_call.return_value = requests.Response()

        kwargs = {}
        r.headers['Proxy-Authenticate'] = 'Basic realm="Fake Realm"'
        guess_auth = GuessProxyAuth(None, None, "user", "passwd")
        guess_auth.handle_407(r, *kwargs)

        proxy_auth_call.assert_called_with(req)



"""Tests for the utils module."""
import io
import os
import os.path

import requests
from requests_toolbelt.downloadutils import stream
from requests_toolbelt.downloadutils import tee
import mock
import pytest

from . import get_betamax


preserve_bytes = {'preserve_exact_body_bytes': True}


def test_stream_response_to_file_uses_content_disposition():
    s = requests.Session()
    recorder = get_betamax(s)
    url = ('https://api.github.com/repos/sigmavirus24/github3.py/releases/'
           'assets/37944')
    filename = 'github3.py-0.7.1-py2.py3-none-any.whl'
    with recorder.use_cassette('stream_response_to_file', **preserve_bytes):
        r = s.get(url, headers={'Accept': 'application/octet-stream'},
                  stream=True)
        stream.stream_response_to_file(r)

    assert os.path.exists(filename)
    os.unlink(filename)


def test_stream_response_to_specific_filename():
    s = requests.Session()
    recorder = get_betamax(s)
    url = ('https://api.github.com/repos/sigmavirus24/github3.py/releases/'
           'assets/37944')
    filename = 'github3.py.whl'
    with recorder.use_cassette('stream_response_to_file', **preserve_bytes):
        r = s.get(url, headers={'Accept': 'application/octet-stream'},
                  stream=True)
        stream.stream_response_to_file(r, path=filename)

    assert os.path.exists(filename)
    os.unlink(filename)


def test_stream_response_to_file_like_object():
    s = requests.Session()
    recorder = get_betamax(s)
    url = ('https://api.github.com/repos/sigmavirus24/github3.py/releases/'
           'assets/37944')
    file_obj = io.BytesIO()
    with recorder.use_cassette('stream_response_to_file', **preserve_bytes):
        r = s.get(url, headers={'Accept': 'application/octet-stream'},
                  stream=True)
        stream.stream_response_to_file(r, path=file_obj)

    assert 0 < file_obj.tell()


@pytest.fixture
def streamed_response(chunks=None):
    chunks = chunks or [b'chunk'] * 8
    response = mock.MagicMock()
    response.raw.stream.return_value = chunks
    return response


def test_tee(streamed_response):
    response = streamed_response
    expected_len = len('chunk') * 8
    fileobject = io.BytesIO()
    assert expected_len == sum(len(c) for c in tee.tee(response, fileobject))
    assert fileobject.getvalue() == b'chunkchunkchunkchunkchunkchunkchunkchunk'


def test_tee_rejects_StringIO():
    fileobject = io.StringIO()
    with pytest.raises(TypeError):
        # The generator needs to be iterated over before the exception will be
        # raised
        sum(len(c) for c in tee.tee(None, fileobject))


def test_tee_to_file(streamed_response):
    response = streamed_response
    expected_len = len('chunk') * 8
    assert expected_len == sum(
        len(c) for c in tee.tee_to_file(response, 'tee.txt')
        )
    assert os.path.exists('tee.txt')
    os.remove('tee.txt')


def test_tee_to_bytearray(streamed_response):
    response = streamed_response
    arr = bytearray()
    expected_arr = bytearray(b'chunk' * 8)
    expected_len = len(expected_arr)
    assert expected_len == sum(
        len(c) for c in tee.tee_to_bytearray(response, arr)
        )
    assert expected_arr == arr


def test_tee_to_bytearray_only_accepts_bytearrays():
    with pytest.raises(TypeError):
        tee.tee_to_bytearray(None, object())



# -*- coding: utf-8 -*-
import mock
import pytest
import requests
import unittest

from requests_toolbelt import SSLAdapter
from . import get_betamax


class TestSSLAdapter(unittest.TestCase):
    def setUp(self):
        self.session = requests.Session()
        self.session.mount('https://', SSLAdapter('SSLv3'))
        self.recorder = get_betamax(self.session)

    def test_klevas(self):
        with self.recorder.use_cassette('klevas_vu_lt_ssl3'):
            r = self.session.get('https://klevas.vu.lt/')
            assert r.status_code == 200

    @pytest.mark.skipif(requests.__build__ < 0x020400,
                        reason="Requires Requests v2.4.0 or later")
    @mock.patch('requests.packages.urllib3.poolmanager.ProxyManager')
    def test_proxies(self, ProxyManager):
        a = SSLAdapter('SSLv3')
        a.proxy_manager_for('http://127.0.0.1:8888')

        assert ProxyManager.call_count == 1
        kwargs = ProxyManager.call_args_list[0][1]
        assert kwargs['ssl_version'] == 'SSLv3'



# -*- coding: utf-8 -*-
import unittest
import sys

from mock import patch
import pytest

from requests_toolbelt.utils import user_agent as ua


class Object(object):
    """
    A simple mock object that can have attributes added to it.
    """
    pass


class TestUserAgent(unittest.TestCase):
    def test_user_agent_provides_package_name(self):
        assert "my-package" in ua.user_agent("my-package", "0.0.1")

    def test_user_agent_provides_package_version(self):
        assert "0.0.1" in ua.user_agent("my-package", "0.0.1")

    def test_user_agent_builds_extras_appropriately(self):
        assert "extra/1.0.0" in ua.user_agent(
            "my-package", "0.0.1", extras=[("extra", "1.0.0")]
        )

    def test_user_agent_checks_extras_for_tuples_of_incorrect_length(self):
        with pytest.raises(ValueError):
            ua.user_agent("my-package", "0.0.1", extras=[
                ("extra", "1.0.0", "oops")
            ])

        with pytest.raises(ValueError):
            ua.user_agent("my-package", "0.0.1", extras=[
                ("extra",)
            ])


class TestImplementationString(unittest.TestCase):
    @patch('platform.python_implementation')
    @patch('platform.python_version')
    def test_cpython_implementation(self, mock_version, mock_implementation):
        mock_implementation.return_value = 'CPython'
        mock_version.return_value = '2.7.5'
        assert 'CPython/2.7.5' == ua._implementation_string()

    @patch('platform.python_implementation')
    def test_pypy_implementation_final(self, mock_implementation):
        mock_implementation.return_value = 'PyPy'
        sys.pypy_version_info = Object()
        sys.pypy_version_info.major = 2
        sys.pypy_version_info.minor = 0
        sys.pypy_version_info.micro = 1
        sys.pypy_version_info.releaselevel = 'final'

        assert 'PyPy/2.0.1' == ua._implementation_string()

    @patch('platform.python_implementation')
    def test_pypy_implementation_non_final(self, mock_implementation):
        mock_implementation.return_value = 'PyPy'
        sys.pypy_version_info = Object()
        sys.pypy_version_info.major = 2
        sys.pypy_version_info.minor = 0
        sys.pypy_version_info.micro = 1
        sys.pypy_version_info.releaselevel = 'beta2'

        assert 'PyPy/2.0.1beta2' == ua._implementation_string()

    @patch('platform.python_implementation')
    def test_unknown_implementation(self, mock_implementation):
        mock_implementation.return_value = "Lukasa'sSuperPython"

        assert "Lukasa'sSuperPython/Unknown" == ua._implementation_string()



"""Module containing the tests for requests_toolbelt.threaded.thread."""
try:
    import queue  # Python 3
except ImportError:
    import Queue as queue
import threading
import unittest
import uuid

import mock
import requests.exceptions

from requests_toolbelt.threaded import thread


def _make_mocks():
    return (mock.MagicMock() for _ in range(4))


def _initialize_a_session_thread(session=None, job_queue=None,
                                 response_queue=None, exception_queue=None):
    with mock.patch.object(threading, 'Thread') as Thread:
        thread_instance = mock.MagicMock()
        Thread.return_value = thread_instance
        st = thread.SessionThread(
            initialized_session=session,
            job_queue=job_queue,
            response_queue=response_queue,
            exception_queue=exception_queue,
            )

    return (st, thread_instance, Thread)


class TestSessionThread(unittest.TestCase):

    """Tests for requests_toolbelt.threaded.thread.SessionThread."""

    def test_thread_initialization(self):
        """Test the way a SessionThread is initialized.

        We want to ensure that we creat a thread with a name generated by the
        uuid module, and that we pass the right method to use as a target.
        """
        with mock.patch.object(uuid, 'uuid4', return_value='test'):
            (st, thread_instance, Thread) = _initialize_a_session_thread()

        Thread.assert_called_once_with(target=st._make_request, name='test')
        assert thread_instance.daemon is True
        assert thread_instance._state is 0
        thread_instance.start.assert_called_once_with()

    def test_is_alive_proxies_to_worker(self):
        """Test that we proxy the is_alive method to the Thread."""
        with mock.patch.object(threading, 'Thread') as Thread:
            thread_instance = mock.MagicMock()
            Thread.return_value = thread_instance
            st = thread.SessionThread(None, None, None, None)

        st.is_alive()
        thread_instance.is_alive.assert_called_once_with()

    def test_join_proxies_to_worker(self):
        """Test that we proxy the join method to the Thread."""
        st, thread_instance, _ = _initialize_a_session_thread()

        st.join()
        thread_instance.join.assert_called_once_with()

    def test_handle_valid_request(self):
        """Test that a response is added to the right queue."""
        session, job_queue, response_queue, exception_queue = _make_mocks()
        response = mock.MagicMock()
        session.request.return_value = response

        st, _, _ = _initialize_a_session_thread(
            session, job_queue, response_queue, exception_queue)

        st._handle_request({'method': 'GET', 'url': 'http://example.com'})
        session.request.assert_called_once_with(
            method='GET',
            url='http://example.com'
        )

        response_queue.put.assert_called_once_with(
            ({'method': 'GET', 'url': 'http://example.com'}, response)
        )
        assert exception_queue.put.called is False
        assert job_queue.get.called is False
        assert job_queue.get_nowait.called is False
        assert job_queue.get_nowait.called is False
        assert job_queue.task_done.called is True

    def test_handle_invalid_request(self):
        """Test that exceptions from requests are added to the right queue."""
        session, job_queue, response_queue, exception_queue = _make_mocks()
        exception = requests.exceptions.InvalidURL()

        def _side_effect(*args, **kwargs):
            raise exception

        # Make the request raise an exception
        session.request.side_effect = _side_effect

        st, _, _ = _initialize_a_session_thread(
            session, job_queue, response_queue, exception_queue)

        st._handle_request({'method': 'GET', 'url': 'http://example.com'})
        session.request.assert_called_once_with(
            method='GET',
            url='http://example.com'
        )

        exception_queue.put.assert_called_once_with(
            ({'method': 'GET', 'url': 'http://example.com'}, exception)
        )
        assert response_queue.put.called is False
        assert job_queue.get.called is False
        assert job_queue.get_nowait.called is False
        assert job_queue.get_nowait.called is False
        assert job_queue.task_done.called is True

    def test_make_request(self):
        """Test that _make_request exits when the queue is Empty."""
        job_queue = next(_make_mocks())
        job_queue.get_nowait.side_effect = queue.Empty()

        st, _, _ = _initialize_a_session_thread(job_queue=job_queue)
        st._make_request()

        job_queue.get_nowait.assert_called_once_with()



"""Module containing the tests for requests_toolbelt.threaded.pool."""
try:
    import queue  # Python 3
except ImportError:
    import Queue as queue
import unittest

import mock
import pytest

from requests_toolbelt.threaded import pool
from requests_toolbelt.threaded import thread


class TestPool(unittest.TestCase):

    """Collection of tests for requests_toolbelt.threaded.pool.Pool."""

    def test_requires_positive_number_of_processes(self):
        """Show that the number of processes has to be > 0."""
        with pytest.raises(ValueError):
            pool.Pool(None, num_processes=0)

        with pytest.raises(ValueError):
            pool.Pool(None, num_processes=-1)

    def test_number_of_processes_can_be_arbitrary(self):
        """Show that the number of processes can be set."""
        p = pool.Pool(None, num_processes=100)
        assert p._processes == 100
        assert len(p._pool) == 100

        p = pool.Pool(None, num_processes=1)
        assert p._processes == 1
        assert len(p._pool) == 1

    def test_initializer_is_called(self):
        """Ensure that the initializer function is called."""
        initializer = mock.MagicMock()
        pool.Pool(None, num_processes=1, initializer=initializer)
        assert initializer.called is True
        initializer.assert_called_once_with(mock.ANY)

    def test_auth_generator_is_called(self):
        """Ensure that the auth_generator function is called."""
        auth_generator = mock.MagicMock()
        pool.Pool(None, num_processes=1, auth_generator=auth_generator)
        assert auth_generator.called is True
        auth_generator.assert_called_once_with(mock.ANY)

    def test_session_is_called(self):
        """Ensure that the session function is called."""
        session = mock.MagicMock()
        pool.Pool(None, num_processes=1, session=session)
        assert session.called is True
        session.assert_called_once_with()

    def test_from_exceptions_populates_a_queue(self):
        """Ensure a Queue is properly populated from exceptions."""
        urls = ["https://httpbin.org/get?n={0}".format(n) for n in range(5)]
        Exc = pool.ThreadException
        excs = (Exc({'method': 'GET', 'url': url}, None) for url in urls)

        job_queue = mock.MagicMock()
        with mock.patch.object(queue, 'Queue', return_value=job_queue):
            with mock.patch.object(thread, 'SessionThread'):
                pool.Pool.from_exceptions(excs)

        assert job_queue.put.call_count == 5
        assert job_queue.put.mock_calls == [
            mock.call({'method': 'GET', 'url': url})
            for url in urls
        ]

    def test_from_urls_constructs_get_requests(self):
        """Ensure a Queue is properly populated from an iterable of urls."""
        urls = ["https://httpbin.org/get?n={0}".format(n) for n in range(5)]

        job_queue = mock.MagicMock()
        with mock.patch.object(queue, 'Queue', return_value=job_queue):
            with mock.patch.object(thread, 'SessionThread'):
                pool.Pool.from_urls(urls)

        assert job_queue.put.call_count == 5
        assert job_queue.put.mock_calls == [
            mock.call({'method': 'GET', 'url': url})
            for url in urls
        ]

    def test_from_urls_constructs_get_requests_with_kwargs(self):
        """Ensure a Queue is properly populated from an iterable of urls."""
        def merge(*args):
            final = {}
            for d in args:
                final.update(d)
            return final

        urls = ["https://httpbin.org/get?n={0}".format(n) for n in range(5)]

        kwargs = {'stream': True, 'headers': {'Accept': 'application/json'}}
        job_queue = mock.MagicMock()
        with mock.patch.object(queue, 'Queue', return_value=job_queue):
            with mock.patch.object(thread, 'SessionThread'):
                pool.Pool.from_urls(urls, kwargs)

        assert job_queue.put.call_count == 5
        assert job_queue.put.mock_calls == [
            mock.call(merge({'method': 'GET', 'url': url}, kwargs))
            for url in urls
        ]

    def test_join_all(self):
        """Ensure that all threads are joined properly."""
        session_threads = []

        def _side_effect(*args, **kwargs):
            thread = mock.MagicMock()
            session_threads.append(thread)
            return thread

        with mock.patch.object(thread, 'SessionThread',
                               side_effect=_side_effect):
            pool.Pool(None).join_all()

        for st in session_threads:
            st.join.assert_called_once_with()

    def test_get_response_returns_thread_response(self):
        """Ensure that a ThreadResponse is made when there's data."""
        queues = []

        def _side_effect():
            q = mock.MagicMock()
            q.get_nowait.return_value = ({}, None)
            queues.append(q)
            return q

        with mock.patch.object(queue, 'Queue', side_effect=_side_effect):
            with mock.patch.object(thread, 'SessionThread'):
                p = pool.Pool(None)

        assert len(queues) == 2

        assert isinstance(p.get_response(), pool.ThreadResponse)
        assert len([q for q in queues if q.get_nowait.called]) == 1

    def test_get_exception_returns_thread_exception(self):
        """Ensure that a ThreadException is made when there's data."""
        queues = []

        def _side_effect():
            q = mock.MagicMock()
            q.get_nowait.return_value = ({}, None)
            queues.append(q)
            return q

        with mock.patch.object(queue, 'Queue', side_effect=_side_effect):
            with mock.patch.object(thread, 'SessionThread'):
                p = pool.Pool(None)

        assert len(queues) == 2

        assert isinstance(p.get_exception(), pool.ThreadException)
        assert len([q for q in queues if q.get_nowait.called]) == 1

    def test_get_response_returns_none_when_queue_is_empty(self):
        """Ensure that None is returned when the response Queue is empty."""
        queues = []

        def _side_effect():
            q = mock.MagicMock()
            q.get_nowait.side_effect = queue.Empty()
            queues.append(q)
            return q

        with mock.patch.object(queue, 'Queue', side_effect=_side_effect):
            with mock.patch.object(thread, 'SessionThread'):
                p = pool.Pool(None)

        assert len(queues) == 2

        assert p.get_response() is None
        assert len([q for q in queues if q.get_nowait.called]) == 1

    def test_get_exception_returns_none_when_queue_is_empty(self):
        """Ensure that None is returned when the exception Queue is empty."""
        queues = []

        def _side_effect():
            q = mock.MagicMock()
            q.get_nowait.side_effect = queue.Empty()
            queues.append(q)
            return q

        with mock.patch.object(queue, 'Queue', side_effect=_side_effect):
            with mock.patch.object(thread, 'SessionThread'):
                p = pool.Pool(None)

        assert len(queues) == 2

        assert p.get_exception() is None
        assert len([q for q in queues if q.get_nowait.called]) == 1

    def test_lists_are_correctly_returned(self):
        """Ensure that exceptions and responses return correct lists."""
        def _make_queue():
            q = queue.Queue()
            q.put(({}, None))
            return q

        with mock.patch.object(thread, 'SessionThread'):
            p = pool.Pool(None)

        # Set up real queues.
        p._response_queue = _make_queue()
        p._exc_queue = _make_queue()

        excs = list(p.exceptions())
        assert len(excs) == 1
        for exc in excs:
            assert isinstance(exc, pool.ThreadException)

        resps = list(p.responses())
        assert len(resps) == 1
        for resp in resps:
            assert isinstance(resp, pool.ThreadResponse)



# -*- coding: utf-8 -*-
import math
import unittest
from requests_toolbelt.multipart.encoder import (
    IDENTITY, MultipartEncoder, MultipartEncoderMonitor
    )


class TestMultipartEncoderMonitor(unittest.TestCase):
    def setUp(self):
        self.fields = {'a': 'b'}
        self.boundary = 'thisisaboundary'
        self.encoder = MultipartEncoder(self.fields, self.boundary)
        self.monitor = MultipartEncoderMonitor(self.encoder)

    def test_content_type(self):
        assert self.monitor.content_type == self.encoder.content_type

    def test_length(self):
        assert self.encoder.len == self.monitor.len

    def test_read(self):
        new_encoder = MultipartEncoder(self.fields, self.boundary)
        assert new_encoder.read() == self.monitor.read()

    def test_callback_called_when_reading_everything(self):
        callback = Callback(self.monitor)
        self.monitor.callback = callback
        self.monitor.read()
        assert callback.called == 1

    def test_callback(self):
        callback = Callback(self.monitor)
        self.monitor.callback = callback
        chunk_size = int(math.ceil(self.encoder.len / 4.0))
        while self.monitor.read(chunk_size):
            pass
        assert callback.called == 5

    def test_bytes_read(self):
        bytes_to_read = self.encoder.len
        self.monitor.read()
        assert self.monitor.bytes_read == bytes_to_read

    def test_default_callable_is_the_identity(self):
        assert self.monitor.callback == IDENTITY
        assert IDENTITY(1) == 1

    def test_from_fields(self):
        monitor = MultipartEncoderMonitor.from_fields(
            self.fields, self.boundary
            )
        assert isinstance(monitor, MultipartEncoderMonitor)
        assert isinstance(monitor.encoder, MultipartEncoder)
        assert monitor.encoder.boundary_value == self.boundary


class Callback(object):
    def __init__(self, monitor):
        self.called = 0
        self.monitor = monitor

    def __call__(self, monitor):
        self.called += 1
        assert monitor == self.monitor



# -*- coding: utf-8 -*-
"""The App Engine Transport Adapter for requests.

This requires a version of requests >= 2.10.0.
"""
import requests
from requests import adapters
from requests import sessions

from .. import exceptions as exc
from .._compat import gaecontrib
from .._compat import timeout

"""
.. versionadded:: 0.6.0

There are two ways to use this library.

If you're using requests directly, you can use code like:
    >>> import requests
    >>> import ssl
    >>> from requests.packages.urllib3.contrib import appengine as ul_appengine
    >>> from requests_toolbelt.adapters import appengine
    >>> s = requests.Session()
    >>> if ul_appengine.is_appengine_sandbox():
    ...    s.mount('http://', appengine.AppEngineAdapter())
    ...    s.mount('https://', appengine.AppEngineAdapter())

If you depend on external libraries which use requests, you can use code like:
    >>> from requests_toolbelt.adapters import appengine
    >>> appengine.monkeypatch()
which will ensure all requests.Session objects use AppEngineAdapter properly.
"""


class AppEngineAdapter(adapters.HTTPAdapter):
    """A transport adapter for Requests to use urllib3's GAE support.

    Implements request's HTTPAdapter API.

    When deploying to Google's App Engine service, some of Requests'
    functionality is broken. There is underlying support for GAE in urllib3.
    This functionality, however, is opt-in and needs to be enabled explicitly
    for Requests to be able to use it.
    """

    def __init__(self, validate_certificate=True, *args, **kwargs):
        _check_version()
        self._validate_certificate = validate_certificate
        super(AppEngineAdapter, self).__init__(self, *args, **kwargs)

    def init_poolmanager(self, connections, maxsize, block=False):
        self.poolmanager = _AppEnginePoolManager(self._validate_certificate)


class _AppEnginePoolManager(object):
    """Implements urllib3's PoolManager API expected by requests.

    While a real PoolManager map hostnames to reusable Connections,
    AppEngine has no concept of a reusable connection to a host.
    So instead, this class constructs a small Connection per request,
    that is returned to the Adapter and used to access the URL.
    """

    def __init__(self, validate_certificate=True):
        self.appengine_manager = gaecontrib.AppEngineManager(
            validate_certificate=validate_certificate)

    def connection_from_url(self, url):
        return _AppEngineConnection(self.appengine_manager, url)

    def clear(self):
        pass


class _AppEngineConnection(object):
    """Implements urllib3's HTTPConnectionPool API's urlopen().

    This Connection's urlopen() is called with a host-relative path,
    so in order to properly support opening the URL, we need to store
    the full URL when this Connection is constructed from the PoolManager.

    This code wraps AppEngineManager.urlopen(), which exposes a different
    API than in the original urllib3 urlopen(), and thus needs this adapter.
    """

    def __init__(self, appengine_manager, url):
        self.appengine_manager = appengine_manager
        self.url = url

    def urlopen(self, method, url, body=None, headers=None, retries=None,
                redirect=True, assert_same_host=True,
                timeout=timeout.Timeout.DEFAULT_TIMEOUT,
                pool_timeout=None, release_conn=None, **response_kw):
        # This function's url argument is a host-relative URL,
        # but the AppEngineManager expects an absolute URL.
        # So we saved out the self.url when the AppEngineConnection
        # was constructed, which we then can use down below instead.

        # We once tried to verify our assumptions here, but sometimes the
        # passed-in URL differs on url fragments, or "http://a.com" vs "/".

        # Jump through the hoops necessary to call AppEngineManager's API.
        return self.appengine_manager.urlopen(
            method,
            self.url,
            body=body,
            headers=headers,
            retries=retries,
            redirect=redirect,
            timeout=timeout,
            **response_kw)


def monkeypatch():
    """Sets up all Sessions to use AppEngineAdapter by default.

    If you don't want to deal with configuring your own Sessions,
    or if you use libraries that use requests directly (ie requests.post),
    then you may prefer to monkeypatch and auto-configure all Sessions.
    """
    _check_version()
    # HACK: We should consider modifying urllib3 to support this cleanly,
    # so that we can set a module-level variable in the sessions module,
    # instead of overriding an imported HTTPAdapter as is done here.
    sessions.HTTPAdapter = AppEngineAdapter


def _check_version():
    if gaecontrib is None:
        raise exc.VersionMismatchError(
            "The toolbelt requires at least Requests 2.10.0 to be "
            "installed. Version {0} was found instead.".format(
                requests.__version__
            )
        )



import requests
from requests.auth import HTTPBasicAuth
from requests_toolbelt.auth.handler import AuthHandler
from requests_toolbelt.auth.handler import NullAuthStrategy


def test_turns_tuples_into_basic_auth():
    a = AuthHandler({'http://example.com': ('foo', 'bar')})
    strategy = a.get_strategy_for('http://example.com')
    assert not isinstance(strategy, NullAuthStrategy)
    assert isinstance(strategy, HTTPBasicAuth)


def test_uses_null_strategy_for_non_matching_domains():
    a = AuthHandler({'http://api.example.com': ('foo', 'bar')})
    strategy = a.get_strategy_for('http://example.com')
    assert isinstance(strategy, NullAuthStrategy)


def test_normalizes_domain_keys():
    a = AuthHandler({'https://API.github.COM': ('foo', 'bar')})
    assert 'https://api.github.com' in a.strategies
    assert 'https://API.github.COM' not in a.strategies


def test_can_add_new_strategies():
    a = AuthHandler({'https://example.com': ('foo', 'bar')})
    a.add_strategy('https://api.github.com', ('fiz', 'baz'))
    assert isinstance(
        a.get_strategy_for('https://api.github.com'),
        HTTPBasicAuth
        )


def test_prepares_auth_correctly():
    # Set up our Session and AuthHandler
    auth = AuthHandler({
        'https://api.example.com': ('bar', 'baz'),
        'https://httpbin.org': ('biz', 'fiz'),
    })
    s = requests.Session()
    s.auth = auth
    # Set up a valid GET request to https://api.example.com/users
    r1 = requests.Request('GET', 'https://api.example.com/users')
    p1 = s.prepare_request(r1)
    assert p1.headers['Authorization'] == 'Basic YmFyOmJheg=='

    # Set up a valid POST request to https://httpbin.org/post
    r2 = requests.Request('POST', 'https://httpbin.org/post', data='foo')
    p2 = s.prepare_request(r2)
    assert p2.headers['Authorization'] == 'Basic Yml6OmZpeg=='

    # Set up an *invalid* OPTIONS request to http://api.example.com
    # NOTE(sigmavirus24): This is not because of the verb but instead because
    # it is the wrong URI scheme.
    r3 = requests.Request('OPTIONS', 'http://api.example.com/projects')
    p3 = s.prepare_request(r3)
    assert p3.headers.get('Authorization') is None



import io

from requests_toolbelt.streaming_iterator import StreamingIterator

import pytest

@pytest.fixture(params=[True, False])
def get_iterable(request):
    '''
    When this fixture is used, the test is run twice -- once with the iterable
    being a file-like object, once being an iterator.
    '''
    is_file = request.param
    def inner(chunks):
        if is_file:
            return io.BytesIO(b''.join(chunks))
        return iter(chunks)
    return inner


class TestStreamingIterator(object):
    @pytest.fixture(autouse=True)
    def setup(self, get_iterable):
        self.chunks = [b'here', b'are', b'some', b'chunks']
        self.size = 17
        self.uploader = StreamingIterator(self.size, get_iterable(self.chunks))

    def test_read_returns_all_chunks_in_one(self):
        assert self.uploader.read() == b''.join(self.chunks)

    def test_read_returns_empty_string_after_exhausting_the_iterator(self):
        for i in range(0, 4):
            self.uploader.read(8192)

        assert self.uploader.read() == b''
        assert self.uploader.read(8192) == b''


class TestStreamingIteratorWithLargeChunks(object):
    @pytest.fixture(autouse=True)
    def setup(self, get_iterable):
        self.letters = [b'a', b'b', b'c', b'd', b'e']
        self.chunks = (letter * 2000 for letter in self.letters)
        self.size = 5 * 2000
        self.uploader = StreamingIterator(self.size, get_iterable(self.chunks))

    def test_returns_the_amount_requested(self):
        chunk_size = 1000
        bytes_read = 0
        while True:
            b = self.uploader.read(chunk_size)
            if not b:
                break
            assert len(b) == chunk_size
            bytes_read += len(b)

        assert bytes_read == self.size

    def test_returns_all_of_the_bytes(self):
        chunk_size = 8192
        bytes_read = 0
        while True:
            b = self.uploader.read(chunk_size)
            if not b:
                break
            bytes_read += len(b)

        assert bytes_read == self.size
